import collections
from ..validators import *
from ..simplifiers import *


class Field(object):
    validator = None
    simplifier = None

    def __init__(self, required=True, is_list=False, is_set=False,
                 *args, **kwargs):
        self.required = required
        self.is_list = is_list
        self.is_set = is_set

    def _validate(self, data):
        return self.validator(data)

    def _validate_list(self, data):
        if not isinstance(data, collections.Iterable):
            raise ValidationError("Is not iterable!")
        return tuple(self._validate(x) for x in data)

    def _validate_set(self, data):
        if not isinstance(data, set) and not isinstance(data, frozenset):
            raise ValidationError("Is not set!")
        return frozenset(self._validate(x) for x in data)

    def validate(self, data, key=None):
        try:
            if self.is_list:
                return self._validate_list(data)
            if self.is_set:
                return self._validate_set(data)
            return self._validate(data)
        except ValidationError as exc:
            if not exc.value:
                exc.value = repr(data)
            exc.path.append(key)
            exc.class_ = self.__class__
            raise exc

    def serialize(self, data):
        if self.is_empty(data):
            return None
        if self.is_list or self.is_set:
            return tuple(self.simplifier.serialize(item) for item in data)
        return self.simplifier.serialize(data)

    def deserialize(self, serialized):
        if serialized is None:
            return self.empty_value()
        if self.is_list:
            return tuple(self.simplifier.deserialize(item) for item in serialized)
        if self.is_set:
            return frozenset(self.simplifier.deserialize(item) for item in serialized)
        return self.simplifier.deserialize(serialized)

    def is_empty(self, value):
        """
        Check value for being empty
        """
        if value is None:
            return True

    def empty_value(self):
        """
        What to return when value has not been set
        """
        if self.is_list:
            return tuple()
        if self.is_set:
            return frozenset()
        return None


class FieldAsIs(Field):
    simplifier = staticmethod(NullSimplifier)


class Bool(FieldAsIs):
    def _validate(self, data):
        if isinstance(data, (bool, int)):
            return bool(data)
        if isinstance(data, str):
            if data.lower() in ('false', '0'):
                return False
            if data.lower() in ('true', '1'):
                return True
        raise ValidationError("Only ints, booleans and strings 'False'' and 'True' are accepted")


class Int(FieldAsIs):
    validator = staticmethod(IntValidator)

    def _validate(self, data):
        if isinstance(data, float):
            raise ValidationError("Floats not allowed")
        return self.validator(data)


class String(FieldAsIs):
    validator = staticmethod(StringValidator)


# as String, but allows int and long
# need it to workaround inconsitent data
class StringInt(FieldAsIs):
    validator = staticmethod(StringIntValidator)


# as String, but allows int, long and float
# need it to workaround inconsitent data
class StringNum(FieldAsIs):
    validator = staticmethod(StringNumValidator)


class Decimal(Field):
    validator = staticmethod(DecimalValidator)
    simplifier = staticmethod(DecimalSimplifier)


class Float(FieldAsIs):
    validator = staticmethod(SimpleTypeValidator(float))


class ViewModelField(Field):
    """
    Validator for a specific StrictDict subclass
    """

    def __init__(self, class_, *args, **kwargs):
        super(ViewModelField, self).__init__(*args, **kwargs)
        self.class_ = class_
        self.simplifier = view_model_simplifier(class_)

    def _validate(self, data):
        if isinstance(data, self.class_):
            return data
        if isinstance(data, collections.Mapping):
            return self.class_(**data)
        raise ValidationError("Not a valid {}!".format(self.class_))


class MapField(Field):
    """
    Validator for a specific StrictDict subclass
    """
    simplifier = staticmethod(NullSimplifier)

    def __init__(self, key_field, value_field, *args, **kwargs):
        super(MapField, self).__init__(*args, **kwargs)
        self.key_field = key_field
        self.value_field = value_field

    def _validate(self, data):
        if not isinstance(data, collections.Mapping):
            raise ValidationError("Not a valid {}!".format(self.class_))
        resp = dict()
        for k, v in data.items():
            key = self.key_field.validate(k)
            value = self.value_field.validate(v)
            resp[key] = value
        return resp


class Date(Field):
    validator = staticmethod(DateValidator)
    simplifier = staticmethod(DateSimplifier)


class DateTime(Field):
    validator = staticmethod(DateTimeValidator)
    simplifier = staticmethod(DateTimeSimplifier)


class Time(Field):
    validator = staticmethod(TimeValidator)
    simplifier = staticmethod(TimeSimplifier)


class TimeStamp(Field):
    validator = staticmethod(TimeStampValidator)
    simplifier = staticmethod(TimeStampSimplifier)



"""
Simplifiers are serializers of sort: they transform (.serialize()) objects to
simpler format understood by undelying storage engine (ints, strings and
dicts mostly) and transform them back (.deserialize()) to python objects, but
perform no checks, so input is supposed to be valid
"""
import datetime as dt
import decimal


class NullSimplifier(object):
    @staticmethod
    def serialize(data):
        return data

    @staticmethod
    def deserialize(data_str):
        return data_str


class DecimalSimplifier(object):
    @staticmethod
    def serialize(data):
        return str(data)

    @staticmethod
    def deserialize(data_str):
        return decimal.Decimal(data_str)


def view_model_simplifier(view_model_class):
    class ViewModelSimplifier(object):
        @staticmethod
        def serialize(obj):
            return obj.simplify()

        @classmethod
        def deserialize(cls, data_str):
            data = view_model_class.restore(data_str)
            return data

    return ViewModelSimplifier


class DateSimplifier(object):
    @staticmethod
    def serialize(data):
        return data.strftime('%Y-%m-%d')

    @classmethod
    def deserialize(cls, data_str):
        data = dt.datetime.strptime(data_str, '%Y-%m-%d').date()
        return data


class DateTimeSimplifier(object):
    @staticmethod
    def serialize(data):
        return data.strftime('%Y-%m-%dT%H:%M')

    @classmethod
    def deserialize(cls, data_str):
        try:
            data = dt.datetime.strptime(data_str, '%Y-%m-%dT%H:%M')
            return data
        except ValueError:
            data = dt.datetime.strptime(data_str, '%Y-%m-%dT%H:%M:%SZ')
            return data


class TimeSimplifier(object):
    @staticmethod
    def serialize(data):
        return data.strftime('%H:%M')

    @classmethod
    def deserialize(cls, data_str):
        try:
            data = dt.datetime.strptime(data_str, '%H:%M').time()
            return data
        except ValueError:
            data = dt.datetime.strptime(data_str, '%H:%M:%S').time()
            return data


class TimeStampSimplifier(object):
    @staticmethod
    def serialize(data):
        return int((data - dt.datetime(1970,1,1)).total_seconds())

    @classmethod
    def deserialize(cls, data_str):
        if isinstance(data_str, int):
            data = dt.datetime.fromtimestamp(data_str)
            return data
        if data_str.isalnum():
            data = dt.datetime.fromtimestamp(int(data_str))
            return data
        try:
            data = dt.datetime.strptime(data_str, '%Y-%m-%dT%H:%M:%S')
            return data
        except ValueError:
            data = dt.datetime.strptime(data_str, '%Y-%m-%dT%H:%M:%SZ')
            return data



"""
Validators check objects on asignment and do coercion if needed
"""
import datetime as dt
import decimal


class ValidationError(Exception):

    def __init__(self, message, class_=None, path=None, value=None, errors=None, *args, **kwargs):
        if errors is None:
            errors = []
        self.errors = errors
        self.path = path or []
        self.value = value
        self.class_ = class_
        self.message = message
        super(ValidationError, self).__init__(*args, **kwargs)

    def __str__(self):
        clsname = self.class_.__name__ if self.class_ else '<No class>'
        path_display = ".".join([str(clsname)] + list(reversed(self.path)))
        return "Failed to validate {0} as {1}: {2}".format(
                self.value, path_display, self.message)


def DummyValidator(data):
    "Passes through whatever comes to it"
    return data


def SimpleTypeValidator(_type, error_classes=[ValueError]):
    def validator(data):
        error_msg = "Failed to validate as %s" % _type
        try:
            return _type(data)
        except tuple(error_classes) as e:
            raise ValidationError("{0}: {1}".format(error_msg, e))
        except TypeError as e:
            raise ValidationError("{0}: {1}".format(error_msg, e))
    return validator


def StringValidator(data):
    if isinstance(data, str):
        return data
    raise ValidationError('Not a string')


def StringIntValidator(data):
    if isinstance(data, str):
        return data
    if isinstance(data, int):
        return str(data)
    raise ValidationError('Not a string')


def StringNumValidator(data):
    if isinstance(data, str):
        return data
    is_num = (isinstance(data, int) or isinstance(data, float))
    if is_num:
        return str(data)
    raise ValidationError('Not a string')


IntValidator = SimpleTypeValidator(int)


def DecimalValidator(data):
    if isinstance(data, float):
        data = str(data)
    return SimpleTypeValidator(
            decimal.Decimal, [decimal.InvalidOperation, ValueError])(data)


def CurrencyValidator(data):
    if not isinstance(data, str):
        raise ValidationError('Not a string')
    if len(data) != 3:
        raise ValidationError('Currency can contain two letters only [%s]' % data)
    return str(data)


def DateTimeValidator(data):
    if isinstance(data, dt.datetime):
        return data
    if isinstance(data, str):
        try:
            return parse_datetime(data)
        except ValueError as e:
            raise ValidationError('Not a datetime [%s, %s]' % (data, e,))
    raise ValidationError('Not a datetime [%s]' % data)


def DateValidator(data):
    if isinstance(data, dt.datetime):
        return data.date()
    if isinstance(data, dt.date):
        return data
    if isinstance(data, str):
        try:
            return parse_date(data)
        except ValueError as e:
            raise ValidationError('Not a date [%s, %s]' % (data, e,))
    raise ValidationError('Not a date [%s]' % data)


def TimeValidator(data):
    if isinstance(data, dt.time):
        return data
    if isinstance(data, str):
        try:
            return parse_time(data)
        except ValueError as e:
            raise ValidationError('Not a time [%s, %s]' % (data, e,))
    raise ValidationError('Not a time [%s]' % data)


def TimeStampValidator(data):
    if isinstance(data, int):
        try:
            return dt.datetime.fromtimestamp(data)
        except (ValueError, TypeError) as exc:
            raise ValidationError('Not a timestamp [%s, %s]' % (data, exc,))
    elif isinstance(data, str):
        try:
            return dt.datetime.fromtimestamp(int(data))
        except (ValueError, TypeError) as exc:
            raise ValidationError('Not a timestamp [%s, %s]' % (data, exc,))
    elif isinstance(data, dt.datetime):
        return data
    raise ValidationError('Not a timestamp [%s]' % data)


def parse_datetime(s):
    parts = s.split('T')
    if len(parts) == 2:
        time_str = parts[1]
        if time_str[-1] == 'Z':
            time_str = time_str[:-1]
        return dt.datetime.combine(parse_date(parts[0]), parse_time(time_str))
    raise ValueError('Invalid datetime format')


def parse_date(s):
    parts = s.split('-')
    if len(parts) == 3:
        return dt.date(int(parts[0]), int(parts[1]), int(parts[2]))
    raise ValueError('Invalid date format')


def parse_time(s):
    parts = s.split(':')
    if len(parts) >= 2:
        s = 0
        mcs = 0
        if len(parts) == 3:
            mparts = parts[2].split('.')
            s = int(mparts[0])
            if len(mparts) > 1:
                mcs = int(mparts[1])
        return dt.time(int(parts[0]), int(parts[1]), s, mcs)
    raise ValueError('Invalid time format')



# coding: utf-8

import pytest
from ..validators import ValidationError
from decimal import Decimal
import datetime as dt
from strictdict import fields as f


def test_string():
    ff = f.String()
    with pytest.raises(ValidationError):
        ff._validate(123)
    result = ff._validate('text')
    assert result == 'text'


def test_decimal():
    ff = f.Decimal()
    result = ff._validate(1870.20)
    assert result == Decimal('1870.2')


def test_time():
    ff = f.Time()
    result = ff._validate('10:11:12')
    assert result == dt.time(10, 11, 12)


def test_bool():
    ff = f.Bool()
    result = ff._validate('false')
    assert result is False
    result = ff._validate('true')
    assert result is True


def test_date():
    ff = f.DateTime()
    result = ff._validate('2013-11-18T00:00:00')
    assert result == dt.datetime(2013, 11, 18)


def test_date_with_z():
    ff = f.DateTime()
    result = ff._validate('2013-11-18T00:00:00Z')
    assert result == dt.datetime(2013, 11, 18)

    result = ff.deserialize("2013-11-18T00:00:00Z")
    assert result == dt.datetime(2013, 11, 18)


def test_timestamp():
    ff = f.TimeStamp()
    now = dt.datetime.now().replace(microsecond=0)
    result = ff._validate(now.strftime('%s'))
    assert result == now
    now_str = now.strftime('%Y-%m-%dT%H:%M:%S')
    result = ff.deserialize(now_str)
    assert result == now
    result = ff.deserialize(now.strftime('%s'))
    assert result == now



# coding: utf-8
"""
Test behaviour of StrictDict mega-class
"""

import pytest

from strictdict import StrictDict
from strictdict import fields as f
from strictdict import ValidationError
from strictdict import api


class Leg(StrictDict):
    is_working = f.Bool()
    is_cool = f.Bool(required=False)
    number = f.Int()
    name = f.String()
    boot_size = f.Float(required=False)
    market_price = f.Decimal(required=False)
    boot_color = f.String(required=False)
    date = f.Date(required=False)


class Centipede(StrictDict):
    age = f.Int()
    legs = f.ViewModelField(class_=Leg, is_list=True)
    favorite_leg = f.ViewModelField(class_=Leg, required=False)


class Missing(object):
    pass


@pytest.fixture(scope='function')
def leg_data():
    return {
        'is_working': True,
        'is_cool': True,
        'number': 1,
        'name': "Martha",
        'boot_color': 'blue',
        'boot_size': 37.5,
        'market_price': '123.45',
    }


@pytest.fixture
def centipede():
    return Centipede(age=100, legs=[leg_data()] * 10, favorite_leg=leg_data())


@pytest.mark.parametrize(('key', 'value'), [
        # OK values
        (None, None),
        ('is_working', 1),
        ('is_working', False),
        ('is_cool', Missing()),
        ('number', 123),
        ('date', '2013-07-08'),
        ('number', '-11'),
        ('boot_color', None),  # OK because boot_color is not required
        ('name', u"Любимая нога дядюшки Лю"),
        ('boot_size', '42.0'),
        ('market_price', 123.45),  # floats ok
        ('market_price', 100500),
])
def test_create_ok(leg_data, key, value):
    if key:
        leg_data[key] = value
    if isinstance(value, Missing):
        del leg_data[key]
    try:
        Leg(**leg_data)
    except ValidationError as exc:
        pytest.fail("{key}={value} raised ValidationError: {exc}".format(
            key=key, value=value, exc=exc))


@pytest.mark.parametrize(('key', 'value'), [
        # Invalid values
        ('number', '123.1'),
        ('number', '-'),
        ('is_working', 'Kinda, yeah, like, whatever'),
        ('boot_size', 'abyr!'),
        ('market_price', 'Too expensive for your cheap ass'),
        # Missing values
        ('is_working', None),
        ('is_working', Missing()),
        ('name', None),
        ('name', Missing()),
        # Extra values
        ('abyr', 'valg'),
        ('glavryba', None)
])
def test_create_invalid(leg_data, key, value):
    if isinstance(value, Missing):
        del leg_data[key]
    else:
        leg_data[key] = value
    with pytest.raises(ValidationError):
        Leg(**leg_data)


@pytest.mark.parametrize(('desc', 'data'), [
    ("Pure dicts",
        {'legs': [leg_data()] * 10, 'favorite_leg': leg_data()}),
    ("Iterable: objects, non-iterable: dict",
     dict(legs=[Leg(**leg_data())] * 10, favorite_leg=leg_data())),
    ("Iterable: empty list, non-iterable: whatever",
        {'legs': [], 'favorite_leg': leg_data()}),
    ("Iterable: objects mixed with dicts, non-iterable: object",
        {'legs': [Leg(**leg_data())] * 5 + [leg_data()] * 3,
         'favorite_leg': Leg(**leg_data())})
])
def test_create_nested_dict(data, desc):
    try:
        Centipede(age=1, **data)
    except ValidationError as exc:
        pytest.fail("Failed in scenario %s: %s" % (desc, exc))


def test_invalid_nested_dict(leg_data):
    valid_leg = leg_data
    invalid_leg = leg_data.copy()
    invalid_leg['is_working'] = '568'

    # Invalid nested
    with pytest.raises(ValidationError):
        Centipede(**{'age': 1,
                   'legs': [valid_leg] * 10,
                   'favorite_leg': invalid_leg})

    # Invalid nested collection
    with pytest.raises(ValidationError):
        Centipede(**{'age': 1,
                   'legs': [valid_leg] * 10 + [invalid_leg],
                   'favorite_leg': valid_leg})

    # Missing required collection
    with pytest.raises(ValidationError):
        Centipede(**{'age': 1,
                    'favorite_leg': valid_leg})


def test_immutability(centipede):
    # Changing scalar attributes
    with pytest.raises(AttributeError):
        centipede.age = 2
    with pytest.raises(AttributeError):
        centipede.legs = []
    with pytest.raises(AttributeError):
        del centipede.age
    # Changing nested object
    with pytest.raises(AttributeError):
        centipede.favorite_leg.name = "Dummy leg"

    with pytest.raises(AttributeError):
        centipede.legs[0].is_working = False
    # Changing iterable length
    with pytest.raises(AttributeError):
        centipede.legs = centipede.legs[1:]

    with pytest.raises(AttributeError):
        centipede.legs.pop()


def test_empty(leg_data):
    class LotsOfEmpty(StrictDict):
        int_ = f.Int(required=False)
        bool_ = f.Bool(required=False)
        leg_ = f.ViewModelField(class_=Leg, required=False)
        strlist = f.String(is_list=True, required=False)

        def empty_checks(self):
            assert self.int_ is None
            assert self.bool_ is None
            assert self.leg_ is None
            assert self.strlist == tuple()

    nothing = LotsOfEmpty()
    nothing.empty_checks()
    # Serialization
    str_ = nothing.to_string()
    nothing_at_all = LotsOfEmpty.loads(str_)
    nothing_at_all.empty_checks()
    # Simplification
    dict_ = nothing.simplify()
    nothing_i_swear = LotsOfEmpty.restore(dict_)
    nothing_i_swear.empty_checks()


def test_to_string(centipede):
    str_ = centipede.to_string()
    restored = Centipede.loads(str_)

    assert centipede.age == restored.age
    assert centipede.legs[1].name == restored.legs[1].name
    assert centipede.favorite_leg.number == restored.favorite_leg.number


def test_dumps(centipede):
    str_ = Centipede.dumps([centipede, centipede])
    c1, c2 = Centipede.loads(str_)

    assert centipede.age == c1.age
    assert centipede.legs[1].name == c1.legs[1].name
    assert centipede.favorite_leg.number == c1.favorite_leg.number
    assert centipede.age == c2.age
    assert centipede.legs[1].name == c2.legs[1].name
    assert centipede.favorite_leg.number == c2.favorite_leg.number


def test_simplify_restore(centipede):
    d = centipede.simplify()
    restored = Centipede.restore(d)

    assert centipede.age == restored.age
    assert centipede.legs[1].name == restored.legs[1].name
    assert centipede.favorite_leg.number == restored.favorite_leg.number


def test_unknown_fields_ignore():
    class Class_v1(StrictDict):
        field1 = f.String(required=True)

    class Class_v2(StrictDict):
        field1 = f.String(required=True)
        field2 = f.String(required=True)

    c2 = Class_v2(field1='f1', field2='f2')
    c2_string = c2.to_string()

    c1 = Class_v1.loads(c2_string)
    c1_dict = dict(**c1)
    assert len(c1_dict) == 1


def test_iteration_for_partial_strict_dict():
    class Class_v1(StrictDict):
        field1 = f.String(required=True)

    class Class_v2(StrictDict):
        field1 = f.String(required=True)
        field2 = f.String(required=True)

    c1 = Class_v1(field1='f1')
    c1_string = c1.to_string()

    c2 = Class_v2.loads(c1_string)
    c2_dict = dict(**c2)
    assert len(c2_dict) == 1


def test_to_dict():
    class SampleInner(StrictDict):
        field1 = f.String(required=True)
        field2 = f.Decimal(required=True)

    class Sample(StrictDict):
        field3 = f.ViewModelField(SampleInner, required=True)
        field4 = f.Decimal(required=True, is_list=True)
        field5 = f.ViewModelField(SampleInner, required=True, is_list=True)

    si = SampleInner(field1='field1', field2=2)
    si2 = SampleInner(field1='field1_2', field2=22)
    s = Sample(field3=si, field4=[1, 2, 3, 4], field5=[si, si2])

    s_dict = s.to_dict()
    assert s_dict['field3']['field1'] == 'field1'
    assert s_dict['field3']['field2'] == 2
    assert s_dict['field4'] == tuple([1, 2, 3, 4])
    assert s_dict['field5'][0]['field1'] == 'field1'
    assert s_dict['field5'][1]['field1'] == 'field1_2'
    assert len(s_dict['field5']) == 2


def test_inheritance():
    class Class_v1(StrictDict):
        __ignored_fields__ = ['field3']
        field1 = f.String(required=True)
        field2 = f.String(required=True)

    class Class_v2(Class_v1):
        field1 = f.Int(required=True)
        field2 = f.String(required=True)
        field3 = f.String(required=True)

    with pytest.raises(ValidationError):
        Class_v2(field1='string1', field2='string2', field3='string3')

    assert 'field3' in Class_v2.__ignored_fields__
    resp = Class_v2(field1=1, field2='string2', field3='string3')
    assert resp
    assert 'field3' not in resp


def test_api():
    class Class_v1(StrictDict):
        field1 = api.ref(f.Int)

    class Class_v2(StrictDict):
        field1 = api.opt(f.Int)

    class Class_v3(StrictDict):
        field1 = api.slist(f.Int)

    class Class_v4(StrictDict):
        field1 = api.optlist(f.Int)

    class Class_v5(StrictDict):
        field1 = api.ref(Class_v1)

    v1 = Class_v1(field1=1)
    assert v1.field1 == 1

    v2 = Class_v2()
    assert v2.field1 is None

    v3 = Class_v3(field1=[1])
    assert isinstance(v3.field1, tuple)
    assert len(v3.field1) == 1
    assert v3.field1[0] == 1

    v4 = Class_v4()
    assert isinstance(v4.field1, tuple)
    assert len(v4.field1) == 0

    v5 = Class_v5(field1=Class_v1(field1=5))
    assert v5.field1.field1 == 5


def test_map_field():
    class Class_v1(StrictDict):
        field1 = api.ref(
            f.MapField,
            api.ref(f.String),
            api.slist(f.Int)
        )

    v1 = Class_v1(field1={'k1': [], 'k2': [1, 2, 3]})
    assert len(v1.field1) == 2
    assert len(v1.field1['k1']) == 0
    assert len(v1.field1['k2']) == 3
    assert v1.field1['k2'][0] == 1
    assert v1.field1['k2'][1] == 2
    assert v1.field1['k2'][2] == 3


def test_load_from_list():
    class Class_v1(StrictDict):
        field = f.Int()

    test_str = '[{"field": 1}, {"field": 2}]'
    res = Class_v1.loads(test_str)
    assert len(res) == 2
    assert res[0].field == 1
    assert res[1].field == 2


def tests_sets():
    class Collector(StrictDict):
        stuff = api.sset(f.String)

    c = Collector(stuff={'mac', 'iphone', 'ipad', 'ipod', 'mac'})
    assert isinstance(c.stuff, frozenset)
    dump = c.to_string()
    restore = Collector.loads(dump)
    assert isinstance(c.stuff, frozenset)
    dump = c.to_string(msg_pack=True)
    restore = Collector.loads(dump, msg_pack=True)
    assert isinstance(c.stuff, frozenset)


def tests_name_collision():
    ex = 'Nothing raised'
    try:
        class BadGuy(StrictDict):
            pop = f.String()
    except Exception as e:
        ex = e
    assert ex.__class__.__name__ == 'NameCollisionError'
